#include <qpixmap.h>
#include <qbuffer.h>
#include <qcstring.h>
#include <qdatetime.h>
#include <qsize.h>
#include <qfile.h>

#include "xmpp.h"
#include "xmpp_tasks.h"
#include "xmpp_xmlcommon.h"
#include "avatars.h"
#include "common.h"
#include "iconset.h"
#include "psiaccount.h"
#include "sha1.h"
#include "base64.h"
#include "profiles.h"

#define AVATAR_SIZE 48

/*
	NOTES:
	- JT_Avatar could move into the Iris backend.

	TODO:
	- Remember bytearray directly as a member
	- Tighten const declarations
*/


/******************************************************************************/

/**
 * Avatar task
 */
class JT_Avatar : public Task
{
	Q_OBJECT

public:
	JT_Avatar(Task*, const Jid& jid);

	void onGo();
	const QString& data()
		{ return data_; }
	const Jid& jid()
		{ return jid_; }
	bool take(const QDomElement&);

private:
	Jid jid_;
	QDomElement iq_;
	QString data_;
};


JT_Avatar::JT_Avatar(Task* parent, const Jid& jid) : Task(parent), jid_(jid)
{
	iq_ = createIQ(doc(), "get", jid_.full(), id());
	QDomElement query = doc()->createElement("query");
	query.setAttribute("xmlns","jabber:iq:avatar");
	iq_.appendChild(query);	
}


void JT_Avatar::onGo()
{
	send(iq_);
}


bool JT_Avatar::take(const QDomElement &x)
{
	if (!iqVerify(x,jid_,id()))
		return false;
	
	if (x.attribute("type") == "result") {
		bool found;
		QDomElement q = queryTag(x);	
		QDomElement tag = findSubTag(q, "data", &found);
		if (found) {
			data_ = tagContent(tag);
		}
		setSuccess();
	}
	else {
		setError(x);
	}

	return true;
}

/******************************************************************************/

/**
 * Avatar: Base class for avatars.
 */
Avatar::Avatar()
{
	mimeType_ = "image/png";
}


Avatar::~Avatar()
{
}


/*const QString& Avatar::hash()
{
	if (hash_.isEmpty()) {
		QByteArray ba;
		QBuffer buffer( ba );
		buffer.open( IO_WriteOnly );
		pixmap().save( &buffer, "PNG" ); 

		// Hash bytearray. TODO: put this in cutestuff ?
		QByteArray hba = SHA1::hash(ba);
		for(int n = 0; n < (int)hba.size(); ++n) {
			QString str;
			str.sprintf("%02x", (uchar)hba[n]);
			hash_.append(str);
		}
	}
	return hash_;
}*/


const QString& Avatar::encode()
{
	if (base64_.isEmpty()) {
		QByteArray ba;
		QBuffer buffer( ba );
		buffer.open( IO_WriteOnly );
		pixmap().save( &buffer, "PNG" ); 
		base64_ = Base64::arrayToString(ba);
	}
	return base64_;
}


const QString& Avatar::mimetype()
{
	return mimeType_;
}


/******************************************************************************/


/**
 * JEP-0008 avatar implementation.
 */
 /* //NO! JEP8 is bad and wrong
class JEP8Avatar : public QObject, public Avatar
{
	Q_OBJECT

public:
	JEP8Avatar(AvatarFactory* factory, const Jid& jid);
	void setHash(const QString& h);

public slots:
	void receiveAvatar();	

signals:
	void avatarChanged(const Jid&);

protected:
	void requestAvatar(const Jid& jid, Task* rootTask);
	bool isCached(const QString& hash);
	void loadFromCache(const QString& hash);
	void saveToCache(const QString& hash, const QByteArray& data);

private:
	Jid jid_;
	QString hash_;
	AvatarFactory* factory_;
};



JEP8Avatar::JEP8Avatar(AvatarFactory* factory, const Jid& jid) : 
	jid_(jid), factory_(factory)
{
}


void JEP8Avatar::setHash(const QString& h)
{
	if (hash_ != h) {
		hash_ = h;
		if (h.isEmpty()) {
			setPixmap(QPixmap());
			emit avatarChanged(jid_);
		}
		else if (isCached(h)) {
			loadFromCache(h);
			emit avatarChanged(jid_);
		}
		else {
			setPixmap(QPixmap());
			requestAvatar(jid_, factory_->account()->client()->rootTask());
		}
	}
}

void JEP8Avatar::requestAvatar(const Jid& jid, Task* rootTask)
{
	printf("Requesting avatar for %s\n",jid.full().latin1());
	JT_Avatar *jav = new JT_Avatar(rootTask, jid);
	connect(jav, SIGNAL(finished()), SLOT(receiveAvatar()));
	jav->go(true);
}

bool JEP8Avatar::isCached(const QString& h)
{
	return QDir(AvatarFactory::getCacheDir()).exists(h);
}

void JEP8Avatar::loadFromCache(const QString& h)
{
	// printf("Loading avatar from cache\n");
	hash_ = h;
	setPixmap(QPixmap(QImage(QDir(AvatarFactory::getCacheDir()).filePath(h))));
	if (pixmap().isNull()) {
		printf("WARNING: Null pixmap. Unsupported format ?\n");
	}
}

void JEP8Avatar::saveToCache(const QString& hash, const QByteArray& data)
{
	// Check if the cache dir exists
	if (!QDir(AvatarFactory::getAvatarsDir()).exists())
		QDir("").mkdir(AvatarFactory::getAvatarsDir());
	if (!QDir(AvatarFactory::getCacheDir()).exists())
		QDir("").mkdir(AvatarFactory::getCacheDir());
	
	// Write file to cache
	// printf("Saving %s to cache.\n",hash.latin1());
	QString fn = QDir(AvatarFactory::getCacheDir()).filePath(hash);
	QFile f(fn);
	if (f.open(IO_WriteOnly)) {
		f.writeBlock(data);
		f.close();
	}
	else
		printf("Error opening %s for writing.\n",f.name().latin1());
	
}

void JEP8Avatar::receiveAvatar()
{
	JT_Avatar *j = (JT_Avatar*) sender();
	if (j->success()) {
		printf("Avatar data arrived from %s\n",j->jid().full().latin1());

		// Strip newlines and whitespace
		QString strippedData(j->data());
		strippedData.remove('\n');

		// Decode data
		QByteArray ba = Base64::stringToArray(strippedData);;
		if (!ba.isEmpty()) {
			saveToCache(hash_,ba);
			setPixmap(QPixmap(ba));
			if (pixmap().isNull()) {
				printf("WARNING: Null pixmap. Unsupported format ?\n");
			}
			emit avatarChanged(j->jid());
		}
		else 
			printf("WARNING: Received data is empty. Bad encoding ?\n");
	}
}*/

/******************************************************************************/

/**
 * FileAvatar: Avatars coming from local files.
 */
class FileAvatar : public Avatar
{
public:
	FileAvatar(const Jid& jid);
	void import(const QString& file);
	QPixmap getPixmap();
	const Jid& getJid() const
		{ return jid_; }

protected:
	bool isDirty() const;
	QString getFileName() const;
	void refresh();
	QDateTime lastModified() const
		{ return lastModified_; }
	
private:
	QDateTime lastModified_;
	Jid jid_;
};


FileAvatar::FileAvatar(const Jid& jid) : jid_(jid)
{
}

void FileAvatar::import(const QString& file)
{
	if (QFileInfo(file).exists()) {
		// Check if the manual dir exists
		if (!QDir(AvatarFactory::getAvatarsDir()).exists())
			QDir("").mkdir(AvatarFactory::getAvatarsDir());
		if (!QDir(AvatarFactory::getManualDir()).exists())
			QDir("").mkdir(AvatarFactory::getManualDir());

		QFile source_file(file);
		QFile target_file(getFileName());
		if (source_file.open(IO_ReadOnly) && target_file.open(IO_WriteOnly)) {
			/* Copy all the data */
			char data[4096];
			while (!source_file.atEnd()) {
				Q_LONG read = source_file.readBlock(data,4096);
				target_file.writeBlock(data,read);
			}
		}	
	}
}

QPixmap FileAvatar::getPixmap() 
{
	refresh();
	return pixmap();
}


void FileAvatar::refresh() 
{
	if (isDirty()) {
		if (QFileInfo(getFileName()).exists()) {
			QImage img(getFileName());
			if (img.width() > AVATAR_SIZE || img.height() > AVATAR_SIZE) {
				img = img.smoothScale(AVATAR_SIZE,AVATAR_SIZE,QImage::ScaleMin);
			}
			setPixmap(QPixmap(img));
		}
		else 
			setPixmap(QPixmap());
	}
}


QString FileAvatar::getFileName() const 
{
	QString f = getJid().bare();
	f.replace('@',"_at_");
	return QDir(AvatarFactory::getManualDir()).filePath(f);
}


bool FileAvatar::isDirty() const
{
	return (pixmap().isNull() 
			|| !QFileInfo(getFileName()).exists() 
			|| QFileInfo(getFileName()).lastModified() > lastModified());
}

/******************************************************************************/

/**
 * ClientAvatar: Avatar representing the logo of a specific client.
 */
class ClientAvatar : public Avatar
{
public:
	ClientAvatar(const QString& name);

protected:
	static ClientAvatar getClientAvatar(const QString& client);
	static QMap<QString,ClientAvatar*> client_avatars;
};


ClientAvatar::ClientAvatar(const QString& name)
{
	QPixmap p;
	if (name == "Psi")
		p = IconsetFactory::icon("psi/psi48").pixmap();
	else if (name == "gaim")
		p = IconsetFactory::icon("client/gaim").pixmap();
	else if (name == "gabber")
		p = IconsetFactory::icon("client/gabber").pixmap();
	else if (name == "Gossip")
		p = IconsetFactory::icon("client/gossip").pixmap();
	else if (name == "Exodus")
		p = IconsetFactory::icon("client/exodus").pixmap();
	else if (name == "Pandion")
		p = IconsetFactory::icon("client/pandion").pixmap();
	else if (name == "Nitro")
		p = IconsetFactory::icon("client/nitro").pixmap();
	else if (name == "MSN-t")
		p = IconsetFactory::icon("client/msn").pixmap();
	else if (name == "JIT") 
		p = IconsetFactory::icon("client/icq").pixmap();
	else 
		p = IconsetFactory::icon("client/unknown").pixmap();

	setPixmap(p);
}

/******************************************************************************/

/**
 * Avatar factory
 */
AvatarFactory::AvatarFactory(PsiAccount* pa) : pa_(pa)
{
}


QPixmap AvatarFactory::getAvatar(const Jid& jid, const QString& resource)
{
	// Compute the avatar of the user
	Avatar* av = retrieveAvatar(jid,resource);

	// If the avatar changed since the previous request, notify everybody of this
	if (av != active_avatars_[jid.full()]) {
		active_avatars_[jid.full()] = av;
		emit avatarChanged(jid);
	}

	return av->getPixmap();
}

Avatar* AvatarFactory::retrieveAvatar(const Jid& jid, const QString& resource)
{
	// printf("Retrieving avatar of %s\n", jid.full().latin1());

	// Try finding a file avatar.
	// printf("File avatar\n");
	if (!file_avatars_.contains(jid.bare())) {
		// printf("File avatar not yet loaded\n");
		file_avatars_[jid.bare()] = new FileAvatar(jid);
	}
	// printf("Trying file avatar\n");
	if (!file_avatars_[jid.bare()]->isEmpty())
		return file_avatars_[jid.bare()];

	// Try finding a JEP-8 avatar.
	//We do *not* condone JEP-8 in any way shape or form. 
	//It's been retracted, do not uncomment the following code.
/*	if (!jep8_avatars_.contains(jid.full())) {
		// printf("JEP 8 avatar not yet loaded\n");
		jep8_avatars_[jid.full()] = new JEP8Avatar(this, jid);
		connect(jep8_avatars_[jid.full()],SIGNAL(avatarChanged(const Jid&)),
					this, SIGNAL(updateAvatar(const Jid&)));
	}
	// printf("Trying JEP8 avatar\n");
	if (!jep8_avatars_[jid.full()]->isEmpty())
		return jep8_avatars_[jid.full()];
	*/
	
	// Return the client avatar 
	// FIXME: The name should be derived directly by the iq. Then,
	// this big if-then-else isn't needed anymore.
	// printf("Falling back on client avatar\n");
	QString name;
	if (resource.contains("Psi",false)) 
		name = "Psi";
	else if (resource.contains("gaim",false)) 
		name = "gaim";
	else if (resource.contains("gabber",false)) 
		name = "gabber";
	else if (resource.contains("Gossip",false)) 
		name = "Gossip";
	else if (resource.contains("Exodus",false)) 
		name = "Exodus";
	else if (resource.contains("Pandion",false)) 
		name = "Pandion";
	else if (resource.contains("Nitro",false)) 
		name = "Nitro";
	else if (resource.contains("MSN")) 
		name = "MSN-t";
	else if (resource.contains("JIT")) 
		name = "JIT";
	else 
		name = "unknown";
	
	if (!client_avatars_.contains(name)) {
		// printf("Client avatar not yet loaded\n");
		client_avatars_[name] = new ClientAvatar(name);
	}

	return client_avatars_[name];
}


/*void AvatarFactory::setHash(const Jid& jid, const QString& hash)
{
	if (!jep8_avatars_.contains(jid.full())) {
		jep8_avatars_[jid.full()] = new JEP8Avatar(this, jid);
		connect(jep8_avatars_[jid.full()],SIGNAL(avatarChanged(const Jid&)),
					this, SLOT(updateAvatar(const Jid&)));
	}

#ifdef JEP8_AVATARS
	jep8_avatars_[jid.full()]->setHash(hash);
#endif
}*/


QPixmap AvatarFactory::getSelfAvatar()
{
	return getAvatar(account()->jid(),"");
}

void AvatarFactory::updateAvatar(const Jid& j)
{
	// FIXME: Maybe we should really look up the client version
	getAvatar(j,"");
}

void AvatarFactory::importManualAvatar(const Jid& j, const QString& fileName)
{
	FileAvatar(j).import(fileName);
	emit avatarChanged(j);
}

QString AvatarFactory::getAvatarsDir()
{
	return pathToProfile(activeProfile) + "/avatars";
}

QString AvatarFactory::getCacheDir()
{
	return getAvatarsDir() + "/cache";
}

QString AvatarFactory::getManualDir()
{
	return getAvatarsDir() + "/manual";
}

/******************************************************************************/

#include "avatars.moc"
